# tested

from boa.interop.Neo.Blockchain import GetTransaction
from boa.interop.Neo.Transaction import *
from boa.interop.Neo.Output import *
from boa.interop.Neo.Input import GetInputHash, GetIndex

NEO = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'
GAS = b'\xe7-(iy\xeel\xb1\xb7\xe6]\xfd\xdf\xb2\xe3\x84\x10\x0b\x8d\x14\x8ewX\xdeB\xe4\x16\x8bqy,`'


def Main(operation, txid):

    tx = GetTransaction(txid)

    if not tx:
        return False

    if operation == 'get_hash':
        return tx.Hash

    elif operation == 'get_type':
        return tx.Type

    elif operation == 'get_attrs':
        return tx.Attributes

    elif operation == 'get_inputs':
        return tx.Inputs

    elif operation == 'get_outputs':
        return tx.Outputs

    elif operation == 'get_references':
        return tx.References

    elif operation == 'get_unspent':
        return tx.UnspentCoins

    elif operation == 'get_output_details':
        res = []

        for item in tx.Outputs:
            subres = []
            subres.append(item.Value)
            subres.append(item.AssetId)
            subres.append(item.ScriptHash)
            res.append(subres)

        return res

    elif operation == 'get_reference_details':
        res = []
        refs = tx.References

        for item in refs:
            subres = []
            subres.append(item.Value)
            subres.append(item.AssetId)
            subres.append(item.ScriptHash)
            res.append(subres)

        return res

    # @TODO
    # For some reason, if theres a bunch of iterations
    # in a row, then the last one fails?
#    elif operation == 'get_unspent_details':
#        res2 = []
#        uns = tx.UnspentCoins
#        for item2 in uns:
#            subres = []
#            subres.append(item2.Value)
#            subres.append(item2.AssetId)
#            subres.append(item2.ScriptHash)
#            res2.append(subres)
#
#        return res2

        # @TODO
    # name clash with Input.GetHash and Transaction.GetHash
    elif operation == 'get_input_details':
        res = []
        inputs = tx.Inputs
        input1 = inputs[0]
        inputhash = GetInputHash(input1)
        inputIndex = input1.Index

        return [inputhash, inputIndex]

    return 'unknown operation'



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_binops(self):
        output = Compiler.instance().load('%s/boa_test/example/BinopTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, ['&', 4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, ['|', 4, 3], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 7)

        tx, results, total_ops, engine = TestBuild(out, ['|', 4, 8], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 12)

        tx, results, total_ops, engine = TestBuild(out, ['^', 4, 4], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 0)

        tx, results, total_ops, engine = TestBuild(out, ['^', 4, 2], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)

        tx, results, total_ops, engine = TestBuild(out, ['>>', 16, 2], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(out, ['>>', 16, 0], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 16)

        tx, results, total_ops, engine = TestBuild(out, ['>>', 11, 1], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 5)

        tx, results, total_ops, engine = TestBuild(out, ['<<', 16, 2], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 64)

        tx, results, total_ops, engine = TestBuild(out, ['<<', 16, -2], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 0)

        tx, results, total_ops, engine = TestBuild(out, ['<<', 4, 5], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 128)

        tx, results, total_ops, engine = TestBuild(out, ['%', 16, 2], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 0)

        tx, results, total_ops, engine = TestBuild(out, ['%', 16, 11], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 5)

        tx, results, total_ops, engine = TestBuild(out, ['//', 16, 2], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)

        tx, results, total_ops, engine = TestBuild(out, ['//', 16, 7], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(out, ['/', 16, 7], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(out, ['~', 16, 0], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -17)

        tx, results, total_ops, engine = TestBuild(out, ['~', -3, 0], self.GetWallet1(), '', '07')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)



# tested

from boa.interop.Neo.TransactionType import *


def Main(operation):
    """
        MinerTransaction = b'\x00'
        IssueTransaction = b'\x01'
        ClaimTransaction = b'\x02'
        EnrollmentTransaction = b'\x20'
        VotingTransaction = b'\x24'
        RegisterTransaction = b'\x40'
        ContractTransaction = b'\x80'
        StateTransaction = b'\x90'
        AgencyTransaction = b'\xb0'
        PublishTransaction = b'\xd0'
        InvocationTransaction = b'\xd1'
    """

    if operation == 'miner':
        return MinerTransaction()

    elif operation == 'issue':
        return IssueTransaction()

    elif operation == 'claim':
        return ClaimTransaction()

    elif operation == 'enrollment':
        return EnrollmentTransaction()

    elif operation == 'voting':
        return VotingTransaction()

    elif operation == 'register':
        return RegisterTransaction()

    elif operation == 'contract':
        return ContractTransaction()

    elif operation == 'state':
        return StateTransaction()

    elif operation == 'agency':
        return AgencyTransaction()

    elif operation == 'publish':
        return PublishTransaction()

    elif operation == 'invocation':
        return InvocationTransaction()

    return 'unknown operation'



from boa.interop.Neo.Iterator import *


def Main(testNum):

    items = {
        'a': 1,
        'c': 4,
        'f': 13
    }

    vals = IterCreate(items)

    if testNum == 1:

        while IterNext(vals):

            print("next!")

        print("ok done")

        return True

    if testNum == 2:
        count = 0
        while next(vals):
            count += 1
        return count

    if testNum == 3:
        i = iter(items)
        keys = []
        while next(i):
            keys.append(i.Key)

        return keys

    if testNum == 4:
        i = iter(items)
        values = []
        while next(i):
            values.append(i.Value)

        return values

    if testNum == 5:
        count = 0
        while vals.Next():
            count += 1
        return count

    if testNum == 6:
        keys = []
        while vals.Keys.next():
            keys.append(vals.Value)
        return keys

    return False




class action():

    method_name = None
    event_name = None
    event_args = None

    def __init__(self, blocks):
        self.method_name = blocks[-1].arg
        self.event_args = [instr.arg for instr in blocks[1:-2]]
        self.event_name = self.event_args[0]



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler
import os


class TestContract(BoaTest):

    TEST_SC_OUTPUT = 'boa_test/example/AddTest1.avm'

    @classmethod
    def setUpClass(cls):
        super(TestContract, cls).setUpClass()
        try:
            os.remove(cls.TEST_SC_OUTPUT)
        except Exception as e:
            pass

    def test_compile_1(self):

        try:
            sc = Compiler.load_and_save('%s/boa_test/example/AddTest1.py' % TestContract.dirname)

            expected_output = '%s/%s' % (TestContract.dirname, self.TEST_SC_OUTPUT)

            self.assertTrue(os.path.exists(expected_output))
        except PermissionError:
            pass

    def test_compile_2(self):

        sc = Compiler.load('%s/boa_test/example/AddTest1.py' % TestContract.dirname)

        default_module = sc.default

        to_s = default_module.to_s()

        output = sc.write()

        self.assertIsInstance(output, bytes)

        self.assertTrue(len(output) > 0)



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler
from neo.Settings import settings

settings.USE_DEBUG_STORAGE = True
settings.DEBUG_STORAGE_PATH = './fixtures/debugstorage'

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_storage_find(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/StorageFindTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, ['prefix'], self.GetWallet1(), '', '10')
        self.assertEqual(len(results), 1)
        resItems = results[0].GetArray()
        self.assertEqual(len(resItems), 3)
        self.assertEqual([1, 2, 3], sorted([item.GetBigInteger() for item in resItems]))

        tx, results, total_ops, engine = TestBuild(out, ['neo'], self.GetWallet1(), '', '10')
        self.assertEqual(len(results), 1)
        resItems = results[0].GetArray()
        self.assertEqual(len(resItems), 0)

        tx, results, total_ops, engine = TestBuild(out, ['blah'], self.GetWallet1(), '', '10')
        self.assertEqual(len(results), 1)
        resItems = results[0].GetArray()
        self.assertEqual(len(resItems), 1)
        self.assertEqual(resItems[0].GetString(), 'Hello Storage Find')

        tx, results, total_ops, engine = TestBuild(out, ['prefix1e'], self.GetWallet1(), '', '10')
        self.assertEqual(len(results), 1)
        resItems = results[0].GetArray()
        self.assertEqual(len(resItems), 2)

        tx, results, total_ops, engine = TestBuild(out, ['pre'], self.GetWallet1(), '', '10')
        self.assertEqual(len(results), 1)
        resItems = results[0].GetArray()
        self.assertEqual(len(resItems), 2)

        tx, results, total_ops, engine = TestBuild(out, ['pref'], self.GetWallet1(), '', '10')
        self.assertEqual(len(results), 1)
        resItems = results[0].GetArray()
        self.assertEqual(len(resItems), 3)
        self.assertEqual(sorted(['prefix1euo', 'prefix1e', 'prefix1__osetuh', ]), sorted([item.GetString() for item in resItems]))




class Iterator:

    @property
    def Value(self):
        return IterValue(self)

    @property
    def Values(self):
        return IterValues(self)

    @property
    def Key(self):
        return IterKey(self)

    @property
    def Keys(self):
        return IterKeys(self)

    def Next(self):
        return IterNext(self)


def IterCreate(items):
    pass


def IterNext(items):
    pass


def IterKey(items):
    pass


def IterKeys(items):
    pass


def IterValue(items):
    pass


def IterValues(items):
    pass



# tested

from boa.interop.Neo.Runtime import Notify
from boa.builtins import concat


def Main():

    m = bytearray(b'\x01\x02\x03\x04\x05\x06\x07\x08')

    # this is a test to see if slice notation without specifying an end
    # is functional
#    s1 = m[2:] # it is not

    # without specifying beginning it is:
    s2 = m[:4]

    j = 2
    k = 4

    s3 = m[j:k]

    Notify(s3)

    s4 = m[get_slice_start():get_slice_end()]

    Notify(s4)

    ind = [1, 3, 4, 5]

    s6 = m[get_slice_start():ind[2]]

    Notify(s6)

    res = concat(s6, concat(s4, concat(s2, s3)))

    Notify(res)

    return res


def get_slice_start():

    return 1


def get_slice_end():
    return 6



# tested


def Main():

    m = bytearray(b'\x01\x02\x03\x04\x05\x06\x07\x08')

    # this is a test to see if slice notation without specifying an end
    # is functional
#    s1 = m[2:] # it is not

    # without specifying beginning it is:
    s2 = m[:4]

    return s2



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler
from neo.Settings import settings

settings.USE_DEBUG_STORAGE = True
settings.DEBUG_STORAGE_PATH = './fixtures/debugstorage'

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_enumerators(self):

        output = Compiler.instance().load('%s/boa_test/example/demo/EnumeratorTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

        tx, results, total_ops, engine = TestBuild(out, [2], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        res = [i.GetBigInteger() for i in results[0].GetArray()]
        self.assertEqual(res, [1, 2, 3, 5, 9, 14])

        tx, results, total_ops, engine = TestBuild(out, [3], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        res = [i.GetBigInteger() for i in results[0].GetArray()]
        self.assertEqual(res, [1, 2, 3, 5, 9, 14])

        tx, results, total_ops, engine = TestBuild(out, [4], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        res = [i.GetString() for i in results[0].GetArray()]
        self.assertEqual(res, ['1', '2', '3', '5', '9', '14', 'a', 'b', 'd', 'f'])

        tx, results, total_ops, engine = TestBuild(out, [5], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)

        tx, results, total_ops, engine = TestBuild(out, [6], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 6)

    def test_iterators(self):

        output = Compiler.instance().load('%s/boa_test/example/demo/IteratorTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), True)

        tx, results, total_ops, engine = TestBuild(out, [2], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

        tx, results, total_ops, engine = TestBuild(out, [3], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        res = [i.GetString() for i in results[0].GetArray()]
        self.assertEqual(res, ['a', 'c', 'f'])

        tx, results, total_ops, engine = TestBuild(out, [4], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        res = [i.GetBigInteger() for i in results[0].GetArray()]
        self.assertEqual(res, [1, 4, 13])

        tx, results, total_ops, engine = TestBuild(out, [5], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

        tx, results, total_ops, engine = TestBuild(out, [6], self.GetWallet1(), '02', '01')
        self.assertEqual(len(results), 1)
        res = [i.GetBigInteger() for i in results[0].GetArray()]
        self.assertEqual(res, [1, 4, 13])



from bytecode import Instr, Compare, Label
from boa.code import pyop
from boa.code.pytoken import PyToken
from bytecode import Bytecode
import ast
import astor


class Expression(object):

    updated_blocklist = None
    block = None
    container_method = None
    tokenizer = None

    next = None

    methodnames = None

    ops = None

    def __init__(self, block, tokenizer, container_method):
        self.block = block
        self.tokenizer = tokenizer
        self.container_method = container_method
        self.methodnames = []
        self.next = None

        self.ops = []
        for item in self.block:
            if isinstance(item, Instr):
                self.ops.append(item.opcode)

    def add_method(self, pytoken):
        self.methodnames.append(pytoken.instruction.arg)

    def _reverselists(self):
        indices = []
        for index, instr in enumerate(self.updated_blocklist):
            if not isinstance(instr, Label) and instr.opcode == pyop.BUILD_LIST:
                indices.append(index)
        if len(indices):
            output = []
            for index, instr in enumerate(self.updated_blocklist):
                output.append(instr)
                if index in indices:
                    output.append(Instr("DUP_TOP", lineno=instr.lineno))
                    output.append(Instr('YIELD_VALUE', lineno=instr.lineno))
            self.updated_blocklist = output

    def _checkbytearray(self):
        to_remove = []
        for index, instr in enumerate(self.block):
            if not isinstance(instr, Label) and instr.opcode == pyop.LOAD_GLOBAL and instr.arg == 'bytearray':
                self.methodnames.append('bytearray')
                to_remove.append(instr)
                to_remove.append(self.block[index + 2])
        if len(to_remove):
            self._remove_instructions(to_remove)

    def _check_load_attr(self):
        replaceable_attr_calls = ['append', 'remove', 'reverse', 'keys', 'values', 'has_key', 'IterKey', 'IterValue', 'IterNext', 'next', ]
        for index, instr in enumerate(self.updated_blocklist):
            if not isinstance(instr, Label) and instr.opcode == pyop.LOAD_ATTR:
                if instr.arg in replaceable_attr_calls:
                    instr.opcode = pyop.LOAD_GLOBAL
                else:
                    attr_name = 'Get%s' % instr.arg
                    module_methods = self.container_method.module.methods
                    matches = []
                    for n in module_methods:
                        if attr_name in n.full_name:
                            matches.append(n)

                    # these two are special cases for name collisions between
                    # block.Hash, tx.Hash, and input.Hash
                    if len(matches) == 0:
                        attr_name = 'GetTX%s' % instr.arg
                        for n in module_methods:
                            if attr_name in n.full_name:
                                matches.append(n)
                    if len(matches) == 0:
                        attr_name = 'GetInput%s' % instr.arg
                        for n in module_methods:
                            if attr_name in n.full_name:
                                matches.append(n)

                    # Special case for enumerators/iterator
                    if len(matches) == 0:
                        attr_name = 'Enumerator%s' % instr.arg
                        for n in module_methods:
                            if attr_name in n.full_name:
                                matches.append(n)
                    if len(matches) == 0:
                        attr_name = 'Iter%s' % instr.arg
                        for n in module_methods:
                            if attr_name == n.full_name.split('.')[-1]:
                                matches.append(n)

                    if len(matches) == 0:
                        raise Exception("Could not load attribute %s " % instr.arg)
                    elif len(matches) > 1:
                        raise Exception(
                            "Could not determine attribute to load. Options: %s " % [n.full_name for n in matches])
                    else:
                        instr.opcode = pyop.LOAD_GLOBAL
                        instr.arg = attr_name
                        self.updated_blocklist = self.block[:index + 1] + [Instr('CALL_FUNCTION', 1, lineno=instr.lineno)] + self.block[index + 1:]

    def _check_function_kwargs(self):
        to_remove = []
        for index, instr in enumerate(self.updated_blocklist):
            if not isinstance(instr, Label) and instr.opcode == pyop.CALL_FUNCTION_KW:
                instr.opcode = pyop.CALL_FUNCTION
                to_remove.append(self.updated_blocklist[index - 1])
        if len(to_remove):
            self._remove_instructions(to_remove)

    def _remove_instructions(self, to_remove):
        updated = []
        for item in self.block:
            if item not in to_remove:
                updated.append(item)
        self.updated_blocklist = updated

    def _checkslice(self):
        last = None
        to_del_index = -1
        for index, instr in enumerate(self.block):
            if isinstance(instr, Instr):
                if last == pyop.BUILD_SLICE and instr.opcode == pyop.BINARY_SUBSCR:
                    to_del_index = index
                last = instr.opcode

        if to_del_index > -1:
            self.block.pop(to_del_index)
            self.updated_blocklist = self.block
            self._checkslice()

    def _ast_to_instr(self, astobj, lineno):

        src = astor.to_source(astobj).replace('"""', "'")

        cp = compile(src, filename='<ast>', mode='eval')
        bc = Bytecode.from_code(cp)
        instructions = bc[:-1]
        for instr in instructions:
            instr.lineno = lineno

        # if its calling a method, we don't want it to do `LOAD_NAME`
        if instructions[-1].opcode == pyop.CALL_FUNCTION and instructions[0].opcode == pyop.LOAD_NAME:
            instructions[0] = Instr("LOAD_GLOBAL", arg=instructions[0].arg, lineno=instructions[0].lineno)

        return instructions

    def _check_dictionary_defs(self):
        new_instr = []
        new_instr_ind = -1

        for index, instr in enumerate(self.block):
            if isinstance(instr, Instr) and instr.opcode == pyop.STORE_FAST:
                for dictionary in self.container_method.dictionary_defs:
                    if dictionary.name == instr.arg:
                        new_instr_ind = index
                        for index, item in enumerate(dictionary.keys):
                            # load the value
                            new_instr += self._ast_to_instr(dictionary.values[index], instr.lineno)
                            # load the dict
                            new_instr.append(Instr("LOAD_FAST", arg=instr.arg, lineno=instr.lineno))
                            # load the key
                            new_instr += self._ast_to_instr(dictionary.keys[index], instr.lineno)
                            new_instr.append(Instr("STORE_SUBSCR", lineno=instr.lineno))

                        self.container_method.dictionary_defs.remove(dictionary)

        if new_instr_ind > -1:
            self.updated_blocklist = self.updated_blocklist[0:new_instr_ind + 1] + new_instr + self.updated_blocklist[new_instr_ind + 1:]

    def _checkloops(self):
        if pyop.SETUP_LOOP in self.ops and pyop.GET_ITER in self.ops:

            counter = self.container_method.forloop_counter

            loopcounter_name = 'forloop_counter_%s' % counter
            looplength_name = 'forloop_length_%s' % counter

            iterable = self.block[1].arg
            iterable_name = self.block[-1].arg
            ln = self.block[0].lineno

            if iterable == 'range' or iterable == 'keys' or iterable == 'values' or self.block[2].opcode == pyop.LOAD_ATTR:

                dynamic_iterable_name = 'dynamic_iterable_%s' % counter

                self.container_method.add_to_scope(dynamic_iterable_name)

                get_iter_index = self.ops.index(pyop.GET_ITER)
                load_range_ops = self.block[1:get_iter_index]
                load_range_ops.append(Instr("STORE_FAST", dynamic_iterable_name, lineno=ln))

                loop_exit = self.block[0].arg
                loop_done = self.block[get_iter_index + 2].arg
                loop_start = self.block[get_iter_index + 1]

                instructions = [
                    #                Instr("SETUP_LOOP",loop_exit,lineno=ln),

                    Instr("LOAD_CONST", 0, lineno=ln),
                    Instr("STORE_FAST", arg=loopcounter_name, lineno=ln)
                ] + load_range_ops + [
                    Instr("LOAD_FAST", arg=dynamic_iterable_name, lineno=ln),
                    Instr("LOAD_GLOBAL", arg="len", lineno=ln),
                    Instr("CALL_FUNCTION", arg=1, lineno=ln),
                    Instr("STORE_FAST", arg=looplength_name, lineno=ln),

                    loop_start,
                    #                Instr("FOR_ITER",loop_done,lineno=ln),
                    Instr("LOAD_FAST", arg=loopcounter_name, lineno=ln),
                    Instr("LOAD_FAST", arg=looplength_name, lineno=ln),
                    Instr("COMPARE_OP", arg=Compare.LT, lineno=ln),
                    Instr("POP_JUMP_IF_FALSE", loop_done, lineno=ln),

                    Instr("LOAD_FAST", arg=dynamic_iterable_name, lineno=ln),
                    Instr("LOAD_FAST", arg=loopcounter_name, lineno=ln),
                    Instr("BINARY_SUBSCR", lineno=ln),
                    Instr("STORE_FAST", arg=iterable_name, lineno=ln),

                    Instr("LOAD_FAST", loopcounter_name, lineno=ln),
                    Instr("LOAD_CONST", 1, lineno=ln),
                    Instr("INPLACE_ADD", lineno=ln),
                    Instr("STORE_FAST", loopcounter_name, lineno=ln),

                ]

            else:
                loop_exit = self.block[0].arg
                loop_done = self.block[4].arg
                loop_start = self.block[3]

                instructions = [
                    #                Instr("SETUP_LOOP",loop_exit,lineno=ln),

                    Instr("LOAD_CONST", 0, lineno=ln),
                    Instr("STORE_FAST", arg=loopcounter_name, lineno=ln),
                    Instr("LOAD_FAST", arg=iterable, lineno=ln),
                    Instr("LOAD_GLOBAL", arg="len", lineno=ln),
                    Instr("CALL_FUNCTION", arg=1, lineno=ln),
                    Instr("STORE_FAST", arg=looplength_name, lineno=ln),

                    loop_start,
                    #                Instr("FOR_ITER",loop_done,lineno=ln),
                    Instr("LOAD_FAST", arg=loopcounter_name, lineno=ln),
                    Instr("LOAD_FAST", arg=looplength_name, lineno=ln),
                    Instr("COMPARE_OP", arg=Compare.LT, lineno=ln),
                    Instr("POP_JUMP_IF_FALSE", loop_done, lineno=ln),

                    Instr("LOAD_FAST", arg=iterable, lineno=ln),
                    Instr("LOAD_FAST", arg=loopcounter_name, lineno=ln),
                    Instr("BINARY_SUBSCR", lineno=ln),
                    Instr("STORE_FAST", arg=iterable_name, lineno=ln),

                    Instr("LOAD_FAST", loopcounter_name, lineno=ln),
                    Instr("LOAD_CONST", 1, lineno=ln),
                    Instr("INPLACE_ADD", lineno=ln),
                    Instr("STORE_FAST", loopcounter_name, lineno=ln),

                ]

            self.container_method.add_to_scope(loopcounter_name)
            self.container_method.add_to_scope(looplength_name)
            self.container_method.add_to_scope(iterable_name)

            self.block = self.updated_blocklist = instructions

    def tokenize(self):

        self.updated_blocklist = self.block
        self._check_dictionary_defs()
        self._checkslice()
        self._checkbytearray()
        self._checkloops()
        self._check_load_attr()
        self._check_function_kwargs()
        self._reverselists()

        ln = None
        last_token = None
        for index, instr in enumerate(self.updated_blocklist):
            if isinstance(instr, Instr):
                ln = instr.lineno

            try:
                token = PyToken(instr, self, index, ln)
                token.to_vm(self.tokenizer, last_token)
            except Exception as e:
                cm = self.container_method
                print("ERROR: %s:%s in %s() with msg - %s" % (cm.module.path, cm.start_line_no + ln - 1, cm.name, str(e)))

            last_token = token

    def lookup_method_name(self, index):
        return self.methodnames.pop()



"""
NEX ICO Template
===================================

Author: Thomas Saunders
Email: tom@neonexchange.org

Date: Dec 11 2017

"""
from boa_test.example.demo.nex.txio import get_asset_attachments
from boa_test.example.demo.nex.token import *
from boa_test.example.demo.nex.crowdsale import *
from boa_test.example.demo.nex.nep5 import *
from boa.interop.Neo.Runtime import GetTrigger, CheckWitness
from boa.interop.Neo.TriggerType import Application, Verification
from boa.interop.Neo.Storage import *

ctx = GetContext()
NEP5_METHODS = ['name', 'symbol', 'decimals', 'totalSupply', 'balanceOf', 'transfer', 'transferFrom', 'approve', 'allowance']


def Main(operation, args):
    """

    :param operation: str The name of the operation to perform
    :param args: list A list of arguments along with the operation
    :return:
        bytearray: The result of the operation
    """

    trigger = GetTrigger()

    # This is used in the Verification portion of the contract
    # To determine whether a transfer of system assets ( NEO/Gas) involving
    # This contract's address can proceed
    if trigger == Verification():

        # check if the invoker is the owner of this contract
        is_owner = CheckWitness(TOKEN_OWNER)

        # If owner, proceed
        if is_owner:

            return True

        # Otherwise, we need to lookup the assets and determine
        # If attachments of assets is ok
        attachments = get_asset_attachments()
        return False
#        return can_exchange(ctx,attachments, True)

    elif trigger == Application():

        for op in NEP5_METHODS:
            if operation == op:
                return handle_nep51(ctx, operation, args)

        if operation == 'deploy':
            return deploy()

        elif operation == 'circulation':
            return get_circulation(ctx)

        # the following are handled by crowdsale

        elif operation == 'mintTokens':
            return perform_exchange(ctx)

        elif operation == 'crowdsale_register':
            return kyc_register(ctx, args)

        elif operation == 'crowdsale_status':
            return kyc_status(ctx, args)

        elif operation == 'crowdsale_available':
            return crowdsale_available_amount(ctx)

        elif operation == 'get_attachments':
            return get_asset_attachments()

        elif operation == 'another_op_1':
            return 1 + 1

        elif operation == 'another_op_2':
            return 1 + 2

        elif operation == 'another_op_3':
            return 1 + 3

        elif operation == 'another_op_4':
            return 1 + 4

        elif operation == 'another_op_5':
            return 1 + 5

        return 'unknown operation'

    return False


def deploy():
    """

    :param token: Token The token to deploy
    :return:
        bool: Whether the operation was successful
    """
    if not CheckWitness(TOKEN_OWNER):
        print("Must be owner to deploy")
        return False

    if not Get(ctx, 'initialized'):
        # do deploy logic
        Put(ctx, 'initialized', 1)
        Put(ctx, TOKEN_OWNER, TOKEN_INITIAL_AMOUNT)
        return add_to_circulation(ctx, TOKEN_INITIAL_AMOUNT)

    return False



from boa.code import pyop
from boa.code.method import method as BoaMethod
from boa.code.action import action as BoaAction
from boa.code.appcall import appcall as BoaAppcall

from boa.util import BlockType, get_block_type
from bytecode import UNSET, Bytecode, BasicBlock, ControlFlowGraph, dump_bytecode, Label
from boa.interop import VMOp
import importlib
from logzero import logger
from collections import OrderedDict
import os
import sys
import hashlib
from boa import __version__
import json


class Module(object):

    bc = None
    cfg = None

    blocks = None

    methods = None
    actions = None
    app_call_registrations = None

    path = None

    all_vm_tokens = OrderedDict()

    module_name = ''

    to_import = None

    _extra_instr = None

    _local_methods = None

    @property
    def extra_instructions(self):
        return self._extra_instr

    @property
    def local_methods(self):
        return self._local_methods

    @staticmethod
    def ImportFromBlock(block: BasicBlock, current_file_path):

        mpath = None
        mnames = []

        filepath = os.path.dirname(os.path.abspath(current_file_path))

        sys.path.append(filepath)

        for index, instr in enumerate(block):
            if instr.opcode == pyop.IMPORT_NAME:
                mpath = instr.arg
            elif instr.opcode == pyop.STORE_NAME:
                if instr.arg not in mnames:
                    mnames.append(instr.arg)
            elif instr.opcode == pyop.IMPORT_STAR:
                mnames = ['*']

        pymodule = importlib.import_module(mpath, mpath)
        filename = pymodule.__file__
        return Module(filename, mpath, mnames)

    @property
    def main(self):
        """
        Return the default method in this module.

        :return: the default method in this module
        :rtype: ``boa.code.method.Method``
        """

        for m in self.methods:

            if m.name in ['Main', 'main']:
                return m

        if len(self.methods):
            return self.methods[0]

        return None

    @property
    def orderered_methods(self):
        """
        An ordered list of methods

        :return: A list of ordered methods is this module
        :rtype: list
        """

        oms = []

        self.methods.reverse()

        if self.main:
            oms = [self.main]

        for m in self.methods:

            if m == self.main:
                continue
            oms.append(m)

        return oms

    def method_by_name(self, method_name):
        """
        Look up a method by its name from the module ``methods`` list.
        :param method_name: the name of the method to look up
        :type method_name: str

        :return: the method ( if it is found)
        :rtype: ``boa.code.method.Method``
        """

        #        print("METHODS FOR MODULE: %s " % self.path)
        for m in self.methods:
            if m.full_name == method_name or m.name == method_name:
                return m

        return None

    def __init__(self, path: str, module_name='', to_import=['*']):

        self.path = path
        self.to_import = to_import
        self.module_name = module_name
        self._local_methods = []
        source = open(path, 'rb')

        compiled_source = compile(source.read(), path, 'exec')

        self.bc = Bytecode.from_code(compiled_source)
        self.cfg = ControlFlowGraph.from_bytecode(self.bc)

        source.close()

        self.build()

    def build(self):
        self.blocks = []
        self.methods = []
        self.actions = []
        self.app_call_registrations = []

        for block in self.cfg:
            start_ln = block[0].lineno
            for index, instr in enumerate(block):
                if instr.lineno != start_ln:
                    self.cfg.split_block(block, index)

        self._extra_instr = []
        new_method_blks = []
        for blk in self.cfg:
            type = get_block_type(blk)
            if type == BlockType.MAKE_FUNCTION:
                new_method_blks.append(blk)
            elif type == BlockType.IMPORT_ITEM:
                new_module = Module.ImportFromBlock(blk, self.path)
                if new_module:
                    for method in new_module.methods:
                        self.methods.append(method)
                    for method in new_module.local_methods:
                        self.methods.append(method)
                    self._extra_instr = self._extra_instr + new_module.extra_instructions
            elif type == BlockType.UNKNOWN:
                self._extra_instr.append(blk)
            elif type == BlockType.CALL_FUNCTION:
                self._extra_instr.append(blk)
            elif type == BlockType.ACTION_REG:
                self.actions.append(BoaAction(blk))
            elif type == BlockType.MAKE_CLASS:
                pass
            elif type == BlockType.APPCALL_REG:
                self.app_call_registrations.append(BoaAppcall(blk))
#            else:
#                logger.info("Block type not used:: %s " % type)

        for m in new_method_blks:

            new_method = BoaMethod(self, m, self.module_name, self._extra_instr)
#            if self.to_import == ['*'] or new_method.name in self.to_import:
            #                print("IMPORTING NEW METHOD %s to module %s " % (new_method.name, self.module_name))

            self.methods.append(new_method)
#            else:
#                self._local_methods.append(new_method)

    def write(self):
        """
        Write the current module to a byte string.

        Note that if you are using the ``Compiler.load('path/to/file.py')``, you must
        call ``module.write()`` before any inspection of the module is possible.

        :return: A bytestring of representing the current module
        :rtype: bytes
        """
        self.link_methods()

        return self.write_methods()

    def write_methods(self):
        """
        Write all methods in the current module to a byte string.

        :return: A bytestring of all current methods in this module
        :rtype: bytes
        """

        b_array = bytearray()
        for key, vm_token in self.all_vm_tokens.items():

            b_array.append(vm_token.out_op)

            if vm_token.data is not None and vm_token.vm_op != VMOp.NOP:
                b_array = b_array + vm_token.data

#        self.to_s()
        return b_array

    def link_methods(self):
        """
        Perform linkage of addresses between methods.
        """

        for method in self.methods:
            method.prepare()

        self.all_vm_tokens = OrderedDict()

        address = 0

        for method in self.orderered_methods:

            if not method.is_interop:
                #                print("ADDING METHOD %s " % method.full_name)
                method.address = address

                for key, vmtoken in method.vm_tokens.items():
                    self.all_vm_tokens[address] = vmtoken
                    address += 1

                    if vmtoken.data is not None and vmtoken.vm_op != VMOp.NOP:
                        address += len(vmtoken.data)

                    vmtoken.addr = vmtoken.addr + method.address

        # loop through once to see if there are any local methods to add
#        local_count = 0
#        for key, vmtoken in self.all_vm_tokens.items():
#            if vmtoken.src_method is not None:
#                target_method, is_local = self.method_by_name(vmtoken.target_method)
#                if is_local:
#                    local_count+=1
#        if local_count > 0:
#            return self.link_methods()

        for key, vmtoken in self.all_vm_tokens.items():
            if vmtoken.src_method is not None:

                target_method = self.method_by_name(vmtoken.target_method)
                if target_method:
                    jump_len = target_method.address - vmtoken.addr
                    vmtoken.data = jump_len.to_bytes(2, 'little', signed=True)
                else:
                    raise Exception("Target method %s not found" % vmtoken.target_method)

    def to_s(self):
        """
        this method is used to print the output of the executable in a readable/ tokenized format.
        sample usage:

        >>> from boa.compiler import Compiler
        >>> module = Compiler.load('./boa/tests/src/LambdaTest.py').default
        >>> module.write()
        >>> module.to_s()
        12            3   LOAD_CONST          9                [data]
                      4   STORE_FAST          j                [data]
        22            11  LOAD_FAST           j                [data]
                      17  CALL_FUNCTION       Main.<locals>.q_1 \
                                          [<boa.code.pytoken.PyToken object at 0x10cb53c50>] [data] 22
                      20  STORE_FAST          m                [data]
        24            27  243                 b'\x03\x00'      [data] 3
                      30  LOAD_FAST           m                [data]
                      35  NOP                                  [data]
                      36  241                                  [data]
                      37  242                                  [data]
                      38  RETURN_VALUE                         [data]
        20            49  243                 b'\x03\x00'      [data] 3
                      52  LOAD_FAST           x                [data]
                      57  LOAD_CONST          1                [data]
                      58  BINARY_ADD                           [data]
                      59  NOP                                  [data]
                      60  241                                  [data]
                      61  242                                  [data]
                      62  RETURN_VALUE                         [data]
        """
        # Initialize if needed
        if self.all_vm_tokens is None:
            self.link_methods()

        lineno = 0
        pstart = True

        for i, (key, value) in enumerate(self.all_vm_tokens.items()):
            if value.pytoken:
                pt = value.pytoken
                do_print_line_no = False
                to_label = None
                from_label = '    '

                if pt.lineno != lineno:
                    print("\n")
                    lineno = pt.lineno
                    do_print_line_no = True

                ds = ''
                if value.data is not None:
                    try:
                        ds = int.from_bytes(value.data, 'little', signed=True)
                    except Exception as e:
                        pass
                    if type(ds) is not int and len(ds) < 1:
                        try:
                            ds = value.data.decode('utf-8')
                        except Exception as e:
                            pass

                lno = "{:<10}".format(
                    pt.lineno if do_print_line_no or pstart else '')
                addr = "{:<5}".format(key)
                op = "{:<20}".format(pt.instruction.name)

                # If this is a number, it is likely a custom python opcode, get the name
                if str(pt.pyop).isnumeric():
                    opname = pyop.to_name(int(str(pt.pyop))).replace('HAVE_ARGUMENT', 'STORE_NAME').replace('YIELD_VALUE', 'REVERSE')
                    if opname is not None:
                        op = "{:<20}".format(opname)

                arg = "{:<50}".format(pt.arg_str)
                data = "[data] {:<20}".format(ds)
                print("%s%s%s%s%s%s" % (lno, from_label, addr, op, arg, data))

            pstart = False

    def export_debug(self, output_path):
        """
        this method is used to generate a debug map for NEO debugger
        """

        file_hash = hashlib.md5(open(output_path, 'rb').read()).hexdigest()
        avm_name = os.path.splitext(os.path.basename(output_path))[0]

        json_data = self.generate_debug_json(avm_name, file_hash)

        mapfilename = output_path.replace('.avm', '.debug.json')
        with open(mapfilename, 'w+') as out_file:
            out_file.write(json_data)

    def generate_debug_json(self, avm_name, file_hash):

        # Initialize if needed
        if self.all_vm_tokens is None:
            self.link_methods()

        lineno = 0

        data = {}
        data['avm'] = {'name': avm_name, 'hash': file_hash}
        data['compiler'] = {'name': 'neo-boa', 'version': __version__}

        files = {}
        breakpoints = []
        data['files'] = files

        map = []
        start_ofs = -1
        last_ofs = 0
        fileid = 0
        for i, (key, value) in enumerate(self.all_vm_tokens.items()):
            if value.pytoken:
                pt = value.pytoken

                if pt.file:
                    if pt.file not in files.keys():
                        fileid = len(files.values()) + 1
                        files[pt.file] = fileid
                    else:
                        fileid = files[pt.file]

                if pt.lineno != lineno:
                    if start_ofs >= 0:
                        map.append({'start': start_ofs, 'end': key - 1, 'file': fileid, 'method': pt.method_name,
                                    'line': lineno, 'file_line_no': pt.method_lineno + lineno})
                    start_ofs = key
                    lineno = pt.lineno

                if pt.is_breakpoint:
                    breakpoints.append(start_ofs)

                last_ofs = key

        if last_ofs >= 0:
            map.append({'start': start_ofs, 'end': last_ofs, 'file': fileid, 'line': lineno})

        data['map'] = map
        data['breakpoints'] = breakpoints
        data['files'] = [{'id': val, 'url': os.path.abspath(key)} for key, val in files.items()]
        json_data = json.dumps(data, indent=4)
        return json_data




from boa.builtins import concat


def Main(operation, idx1, idx2):

    print("hello")

    # idx1 = 1
    # idx2 = 3

    mylist = [1, 2, 3, 5, 9, 1000, 32, -1]

    mystr_list = ['ab', 'bc', 'de', 'ef']

    if operation == 'add':

        return mylist[idx1] + mylist[idx2]

    elif operation == 'sub':

        return mylist[idx1] - mylist[idx2]

    elif operation == 'fun':
        # my_method( 4, 6 ) == 10

        return my_method(my_method_2(mylist[idx1]), my_method_3(mylist[idx2]))

#    elif operation == 'concat_fun':

#        return concat(mystr_list[idx1], concat(mystr_list[idx2], concat(mystr_list[idx1])))

    return False


def my_method(a, b):

    return a + b


def my_method_2(c):

    return c * 2


def my_method_3(j):

    return j + 1




def Main(operation):

    mylist = [1, 4, 10]

    q = b'\xab'

    myDict = {
        'a': 1,
        'b': 34,
        'c': 'abc',
        q: '123'
    }

    if operation == 1:

        return 'a' in myDict

    elif operation == 2:

        return 3 in myDict

    elif operation == 3:

        # this should return true
        # the neo-vm uses 'HAS_KEY' op for lists
        # which only checks the length of the array :(
        # it doesn't test whether the array
        # contains the value
        return 2 in mylist

    elif operation == 4:
        # this should return false
        # even though 10 is in the list
        # because the list isn't 10 items long
        return 10 in mylist

    elif operation == 5:

        myDict['h'] = 3

        return 'h' in myDict

    elif operation == 6:

        mylist.append(8)

        return 3 in mylist

    elif operation == 7:

        myDict[7] = 'a'

        return 7 in myDict

    elif operation == 8:

        myDict.remove('a')

        return 'a' in myDict

    elif operation == 9:

        return b'\xab' in myDict

    return False



"""
Basic settings for an NEP5 Token and crowdsale
"""

from boa.interop.Neo.Storage import *

TOKEN_NAME = 'NEX Template V2'

TOKEN_SYMBOL = 'NXT2'

TOKEN_DECIMALS = 8

# This is the script hash of the address for the owner of the token
# This can be found in ``neo-python`` with the wallet open, use ``wallet`` command
TOKEN_OWNER = b'S\xefB\xc8\xdf!^\xbeZ|z\xe8\x01\xcb\xc3\xac/\xacI)'

TOKEN_CIRC_KEY = b'in_circulation'

TOKEN_TOTAL_SUPPLY = 10000000 * 100000000  # 10m total supply * 10^8 ( decimals)

TOKEN_INITIAL_AMOUNT = 2500000 * 100000000  # 2.5m to owners * 10^8

# for now assume 1 dollar per token, and one neo = 40 dollars * 10^8
TOKENS_PER_NEO = 40 * 100000000

# for now assume 1 dollar per token, and one gas = 20 dollars * 10^8
TOKENS_PER_GAS = 20 * 100000000

# maximum amount you can mint in the limited round ( 500 neo/person * 40 Tokens/NEO * 10^8 )
MAX_EXCHANGE_LIMITED_ROUND = 500 * 40 * 100000000

# when to start the crowdsale
BLOCK_SALE_START = 755000

# when to end the initial limited round
LIMITED_ROUND_END = 755000 + 10000

KYC_KEY = b'kyc_ok'

LIMITED_ROUND_KEY = b'r1'


def crowdsale_available_amount(ctx):
    """

    :return: int The amount of tokens left for sale in the crowdsale
    """

    in_circ = Get(ctx, TOKEN_CIRC_KEY)

    available = TOKEN_TOTAL_SUPPLY - in_circ

    return available


def add_to_circulation(ctx, amount):
    """
    Adds an amount of token to circlulation

    :param amount: int the amount to add to circulation
    """

    current_supply = Get(ctx, TOKEN_CIRC_KEY)

    current_supply += amount
    Put(ctx, TOKEN_CIRC_KEY, current_supply)
    return True


def get_circulation(ctx):
    """
    Get the total amount of tokens in circulation

    :return:
        int: Total amount in circulation
    """
    return Get(ctx, TOKEN_CIRC_KEY)



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_Demo(self):

        output = Compiler.instance().load('%s/boa_test/example/demo/Demo1.py' % TestContract.dirname).default.write()

        tx, results, total_ops, engine = TestBuild(output, ['add', 1, 3], self.GetWallet1(), '070202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 7)

        tx, results, total_ops, engine = TestBuild(output, ['add', 2, 3], self.GetWallet1(), '070202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 8)




from boa.builtins import concat


def Main(ba1, ba2):

    m = ba2[1:2]  # but you can do this instead

    # strings and byte arrays work the same
    mystr = 'staoheustnau'

    # this will not work
    # m = mystr[3]

    # but this will
    m = mystr[3:5]

    #
    m = ba1[1:len(ba1)]

    return concat(m, concat(mystr, ba2))



# tested


def Main():

    mylist = [1, 4, 6, 9, 13]

    d2 = add(mylist[0], mylist[2], get_a_value(mylist[1]))

    return d2


def add(a, b, c):

    result = a + b + c

    return result


def get_a_value(m):

    return m + 4



from bytecode import Instr, Bytecode, Label
from boa.code.vmtoken import VMTokenizer
from boa.code.expression import Expression
from boa.code import pyop
from boa.code.ast_preprocess import preprocess_method_body
from uuid import uuid4

import pdb
import dis


class method(object):

    code = None
    bytecode = None

    block = None

    blocks = []

    stack_size = 0

    tokens = []
    tokenizer = None

    address = 0

    module = None

    name = None
    module_name = None

    dictionary_defs = None

    start_line_no = None

    _blocks = None
    _expressions = None

    _scope = None

    _forloop_counter = 0

    _extra = None

    _id = None

    code_object = None

    @property
    def id(self):
        return self._id

    @property
    def forloop_counter(self):
        self._forloop_counter += 1
        return self._forloop_counter

    @property
    def vm_tokens(self):
        """
        Returns a list of all vm tokens in this method.

        :return: a list of vm tokens in this method
        :rtype: list
        """

        return self.tokenizer.vm_tokens

    @property
    def is_interop(self):
        if 'boa.interop' in self.full_name:
            return True
        if 'boa.builtins' in self.full_name and self.full_name != 'boa.builtins.range':
            return True
        return False

    @property
    def full_name(self):
        if len(self.module_name):
            return '%s.%s' % (self.module_name, self.name)
        return self.name

    @property
    def scope(self):
        return self._scope

    @property
    def args(self):
        return self.bytecode.argnames

    @property
    def stacksize(self):
        return self.bytecode.argcount + len(self._blocks) + 2

    def __init__(self, module, block, module_name, extra):
        self.module = module
        self.block = block
        self.module_name = module_name
        self._extra = extra
        self._id = uuid4()
        self.name = self.block[1].arg
        self.start_line_no = self.block[0].lineno
        self.code_object = self.block[0].arg

#        dis.dis(code_object)
        self.code, self.dictionary_defs = preprocess_method_body(self.code_object)

        self.bytecode = Bytecode.from_code(self.code)

        self.setup()

    def setup(self):

        self._scope = {}

        for index, name in enumerate(self.bytecode.argnames):
            self._scope[name] = index

        blocks = []

        # find LOAD_GLOBALS
        gbl = []
        for instr in self.bytecode:
            if isinstance(instr, Instr) and instr.opcode == pyop.LOAD_GLOBAL:
                gbl.append(instr.arg)

        # if there are global things passed in
        # we want to check if they are used in the method
        # and if so, load them in
        global_blocks = []

        if len(self._extra):
            for item in self._extra:
                if item[-1].opcode == pyop.STORE_NAME:
                    if item[-1].arg in gbl:
                        global_blocks.append(item)
                        self.add_to_scope(item[-1].arg)
                        if item[0].opcode == pyop.LOAD_NAME:
                            item[0].opcode = pyop.LOAD_GLOBAL
            blocks = global_blocks

        instructions = []
        last_ln = self.bytecode[0].lineno
        for instr in self.bytecode:
            if not isinstance(instr, Label) and instr.lineno != last_ln:
                last_ln = instr.lineno
                if len(instructions):
                    blocks.append(instructions)
                instructions = []
            if not isinstance(instr, Label) and instr.opcode == pyop.STORE_FAST:
                self.add_to_scope(instr.arg)

            instructions.append(instr)
        if len(instructions):
            blocks.append(instructions)

        self._blocks = blocks

        self.tokenizer = VMTokenizer(self)

        self._expressions = []

    def add_to_scope(self, argname):
        if argname not in self.scope.keys():
            current_total = len(self._scope)
            self._scope[argname] = current_total

    def prepare(self):

        last_exp = None
        for block in self._blocks:
            exp = Expression(block, self.tokenizer, self)
            self._expressions.append(exp)
            if last_exp:
                last_exp.next = exp
            last_exp = exp

        for exp in self._expressions:
            exp.tokenize()

        self.convert_breaks()
        self.convert_jumps()

    def convert_jumps(self):
        filtered = []
        for vmtoken in self.tokenizer.vm_tokens.values():
            if vmtoken.pytoken:
                filtered.append(vmtoken)
        for vmtoken in filtered:
            if vmtoken.pytoken.jump_from and not vmtoken.pytoken.jump_found:
                for vmtoken2 in filtered:
                    if vmtoken2.pytoken.jump_target == vmtoken.pytoken.jump_from:
                        diff = vmtoken2.addr - vmtoken.addr
                        vmtoken.data = diff.to_bytes(2, 'little', signed=True)
                        vmtoken2.pytoken.jump_from_addr = vmtoken.addr
                        vmtoken.pytoken.jump_to_addr = vmtoken2.addr

    def convert_breaks(self):
        tokens = list(self.tokenizer.vm_tokens.values())
        setup_token_label = None

        for tkn in tokens:
            if tkn.pytoken:
                if tkn.pytoken.pyop == pyop.SETUP_LOOP:
                    setup_token_label = tkn.pytoken.jump_from
                if tkn.pytoken.pyop == pyop.BREAK_LOOP:
                    if not setup_token_label:
                        raise Exception("No loopsetup for break")
                    tkn.pytoken.jump_from = setup_token_label




class Enumerator:

    @property
    def Value(self):
        return EnumeratorValue(self)

    def Next(self):
        return EnumeratorNext(self)


def EnumeratorCreate(items):
    pass


def EnumeratorNext(enumerator):
    pass


def EnumeratorValue(enumerator):
    pass


def EnumeratorConcat(enumerator1, enumerator2):
    pass



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler

from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_AddTest(self):

        output = Compiler.instance().load('%s/boa_test/example/AddTest.py' % TestContract.dirname).default.write()

        tx, results, total_ops, engine = TestBuild(output, [2], self.GetWallet1(), '02', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 4)

        tx, results, total_ops, engine = TestBuild(output, [23234], self.GetWallet1(), '02', '02')
        self.assertEqual(results[0].GetBigInteger(), 23236)

        tx, results, total_ops, engine = TestBuild(output, [0], self.GetWallet1(), '02', '02')
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(output, [-112], self.GetWallet1(), '02', '02')
        self.assertEqual(results[0].GetBigInteger(), -110)

    def test_AddTest1(self):

        output = Compiler.instance().load('%s/boa_test/example/AddTest1.py' % TestContract.dirname).default.write()

        tx, results, total_ops, engine = TestBuild(output, [1, 2, 3, 4], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 9)

        tx, results, total_ops, engine = TestBuild(output, [0, 0, 0, 2], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 2)

        tx, results, total_ops, engine = TestBuild(output, [-2, 3, -6, 2], self.GetWallet1(), '02020202', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -2)

    def test_AddTest2(self):

        output = Compiler.instance().load('%s/boa_test/example/AddTest2.py' % TestContract.dirname).default.write()

        tx, results, total_ops, engine = TestBuild(output, [], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), 3)

    def test_AddTest4(self):

        output = Compiler.instance().load('%s/boa_test/example/AddTest4.py' % TestContract.dirname).default.write()

        tx, results, total_ops, engine = TestBuild(output, [1, 2, 3, 4], self.GetWallet1(), '', '02')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBigInteger(), -9)

    def test_AddVoid(self):

        output = Compiler.instance().load('%s/boa_test/example/AddTestVoid.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, [3], self.GetWallet1(), '02', 'ff')
        # this should I guess return nothing
        # for now it returns an empty byte array
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b''))



from boa.interop.Neo.Enumerator import *


def Main(testNum):

    items = [1, 2, 3, 5, 9, 14]

    vals = EnumeratorCreate(items)

    if testNum == 1:

        while EnumeratorNext(vals):

            print("next!")

        print("ok done")

        return True

    if testNum == 2:

        ret = []
        while EnumeratorNext(vals):
            ret.append(vals.Value)

        return ret

    if testNum == 3:
        ret = []
        vals = enumerate(items)

        while EnumeratorNext(vals):
            ret.append(vals.Value)

        return ret

    if testNum == 4:

        items2 = ['a', 'b', 'd', 'f']

        enum2 = EnumeratorCreate(items2)

        doublenumerator = EnumeratorConcat(vals, enum2)

        ret = []

        while EnumeratorNext(doublenumerator):
            ret.append(doublenumerator.Value)
        return ret

    if testNum == 5:

        count = 0
        while next(vals):
            count += 1
        return count

    if testNum == 6:

        myenum = enumerate(items)
        count = 0
        while myenum.next():
            count += 1
        return count

    return False





def Main(operation):

    if operation == 1:
        return operation + 1

    elif operation == 2:
        return operation + 1

    elif operation == 3:
        return operation + 1

    elif operation == 4:
        return operation + 1

    elif operation == 5:
        return operation + 1

    elif operation == 6:
        return operation + 1

    elif operation == 7:
        return operation + 1

    elif operation == 8:
        return operation + 1

    elif operation == 9:
        return operation + 1

    elif operation == 10:
        return operation + 1

    elif operation == 11:
        return operation + 1

    elif operation == 12:
        return operation + 1

    elif operation == 13:
        return operation + 1

    elif operation == 14:
        return operation + 1

    elif operation == 15:
        return operation + 1

    elif operation == 16:
        return operation + 1

    elif operation == 17:
        return operation + 1

    elif operation == 18:
        return operation + 1

    return -1



from boa_test.tests.boa_test import BoaTest
from boa.compiler import Compiler
from neo.VM.InteropService import StackItem, Array, ByteArray
from neocore.IO.BinaryReader import BinaryReader
from neo.IO.MemoryStream import StreamManager
from neo.Prompt.Commands.BuildNRun import TestBuild


class TestContract(BoaTest):

    def test_serialization_1(self):
        output = Compiler.instance().load('%s/boa_test/example/demo/SerializationTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [1], self.GetWallet1(), '02', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x80\x05\x00\x01a\x02\x01\x03\x80\x03\x00\x01j\x02\x01\x03\x02\x01\x05\x00\x02jk\x00\x07lmnopqr'))

    def test_serialization_2(self):
        output = Compiler.instance().load('%s/boa_test/example/demo/SerializationTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [2], self.GetWallet1(), '02', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x80\x05\x00\x01a\x02\x01\x03\x80\x03\x00\x01j\x02\x01\x03\x02\x01\x05\x00\x02jk\x00\x07lmnopqr'))

        stream = StreamManager.GetStream(results[0].GetByteArray())
        reader = BinaryReader(stream)
        stack_item = StackItem.DeserializeStackItem(reader)

        self.assertIsInstance(stack_item, Array)
        self.assertEqual(stack_item.Count, 5)
        self.assertEqual(stack_item.GetArray()[-1].GetString(), 'lmnopqr')

    def test_serialization_3(self):
        output = Compiler.instance().load('%s/boa_test/example/demo/SerializationTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [3], self.GetWallet1(), '02', '05')
        self.assertEqual(len(results), 1)
        stack_item = results[0].GetArray()

        self.assertIsInstance(stack_item, list)
        self.assertEqual(len(stack_item), 5)
        self.assertEqual(stack_item[-1].GetString(), 'lmnopqr')

    def test_serialization_4(self):
        output = Compiler.instance().load('%s/boa_test/example/demo/SerializationTest.py' % TestContract.dirname).default
        out = output.write()
        tx, results, total_ops, engine = TestBuild(out, [4], self.GetWallet1(), '02', '05')
        self.assertEqual(len(results), 1)
        stack_item = results[0].GetArray()

        self.assertIsInstance(stack_item, list)
        self.assertEqual(len(stack_item), 3)
        self.assertEqual(stack_item[-1].GetBigInteger(), 5)




def Main(fibnumber):
    """

    :param fibnumber:
    :return:
    """
    fibresult = fibR(fibnumber)

    return fibresult


def fibR(n):
    """

    :param n:
    :return:
    """
    if n == 1 or n == 2:
        return 1

    return fibR(n - 1) + fibR(n - 2)




def Verification():
    """

    :return:
    """
    pass


def VerificationR():
    """

    :return:
    """
    pass


def Application():
    """

    :return:
    """
    pass


def ApplicationR():
    """

    :return:
    """
    pass



from boa_test.tests.boa_test import BoaFixtureTest
from boa.compiler import Compiler
from neo.Core.TX.Transaction import Transaction, TransactionOutput, TransactionInput
from neo.Core.TX.TransactionAttribute import TransactionAttribute
from neo.Core.CoinReference import CoinReference
from neo.Prompt.Commands.BuildNRun import TestBuild

NEO = bytearray(b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5')
GAS = bytearray(b'\xe7-(iy\xeel\xb1\xb7\xe6]\xfd\xdf\xb2\xe3\x84\x10\x0b\x8d\x14\x8ewX\xdeB\xe4\x16\x8bqy,`')


class TestContract(BoaFixtureTest):

    def test_Output(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/OutputsTest.py' % TestContract.dirname).default
        out = output.write()

        txid = bytearray(b'\xb4A?l#\xdc@7ki<)\x05\xed\xd5\x9a"\xc3I\x10-\x9f#[\xfc\xf6\xb1$N\\\xdb\xce')

        tx, results, total_ops, engine = TestBuild(out, [txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(len(results[0].GetArray()), 2)

    def test_TransactionTypes(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/TransactionTypeTest.py' % TestContract.dirname).default
        out = output.write()

        tx, results, total_ops, engine = TestBuild(out, ['miner'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x00')

        tx, results, total_ops, engine = TestBuild(out, ['issue'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x01')

        tx, results, total_ops, engine = TestBuild(out, ['claim'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x02')

        tx, results, total_ops, engine = TestBuild(out, ['enrollment'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x20')

        tx, results, total_ops, engine = TestBuild(out, ['voting'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x24')

        tx, results, total_ops, engine = TestBuild(out, ['register'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x40')

        tx, results, total_ops, engine = TestBuild(out, ['contract'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x80')

        tx, results, total_ops, engine = TestBuild(out, ['state'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\x90')

        tx, results, total_ops, engine = TestBuild(out, ['agency'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\xb0')

        tx, results, total_ops, engine = TestBuild(out, ['publish'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\xd0')

        tx, results, total_ops, engine = TestBuild(out, ['invocation'], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), b'\xd1')

    def test_Transaction(self):

        output = Compiler.instance().load('%s/boa_test/example/blockchain/TransactionTest.py' % TestContract.dirname).default
        out = output.write()

        txid = bytearray(b'\xb4A?l#\xdc@7ki<)\x05\xed\xd5\x9a"\xc3I\x10-\x9f#[\xfc\xf6\xb1$N\\\xdb\xce')

        bad_tx = bytearray(b'\xb4A?l#\xdc@7ki<)\x05\xed\xd5\x9a"\xc3I\x10-\x9f#[\xfc\xf6\xb1$N\\\xdb\xca')

        tx, results, total_ops, engine = TestBuild(out, ['get_hash', bad_tx], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetBoolean(), False)

        tx, results, total_ops, engine = TestBuild(out, ['get_hash', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), txid)

        tx, results, total_ops, engine = TestBuild(out, ['get_type', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0].GetByteArray(), bytearray(b'\x80'))

        tx, results, total_ops, engine = TestBuild(out, ['get_attrs', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0].GetInterface(), TransactionAttribute)

        tx, results, total_ops, engine = TestBuild(out, ['get_inputs', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0].GetInterface(), CoinReference)

        tx, results, total_ops, engine = TestBuild(out, ['get_outputs', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 2)
        self.assertIsInstance(res[0].GetInterface(), TransactionOutput)

        tx, results, total_ops, engine = TestBuild(out, ['get_references', txid], self.GetWallet1(), '07', '05')
        res = results[0].GetArray()
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0].GetInterface(), TransactionOutput)

        tx, results, total_ops, engine = TestBuild(out, ['get_unspent', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 1)
        self.assertIsInstance(res[0].GetInterface(), TransactionOutput)

        tx, results, total_ops, engine = TestBuild(out, ['get_output_details', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 2)

        o1 = res[0].GetArray()
        self.assertEqual(len(o1), 3)
        self.assertEqual(o1[0].GetBigInteger(), 100000000)
        self.assertEqual(o1[1].GetByteArray(), NEO)
        self.assertEqual(o1[2].GetByteArray(), bytearray(b"\xe5\x8eY\x99\xbc\xbf]x\xf5.\xad@eA1\xab\xb9\xee\'\t"))

        tx, results, total_ops, engine = TestBuild(out, ['get_reference_details', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 1)

        o1 = res[0].GetArray()
        self.assertEqual(len(o1), 3)
        self.assertEqual(o1[0].GetBigInteger(), 98700000000)
        self.assertEqual(o1[1].GetByteArray(), NEO)
        self.assertEqual(o1[2].GetByteArray(), bytearray(b"\xe5:\'\xd3}\x7fZ1\x87\x00<!\xef\xe3rS\x04\xa7A\x06"))

        tx, results, total_ops, engine = TestBuild(out, ['get_input_details', txid], self.GetWallet1(), '07', '05')
        self.assertEqual(len(results), 1)
        res = results[0].GetArray()
        self.assertEqual(len(res), 2)
        self.assertEqual(res[0].GetByteArray(), bytearray(b'\x02\x06\xcco\x91\x96\x95\xfbU\xc9`\\U\x12q(\xc2\x96\x97\xd7\x91\xaf\x88L&6Ali\xa9D\x88'))
        self.assertEqual(res[1].GetBigInteger(), 1)



# tested

from boa.builtins import sha1, sha256, hash160, hash256


def Main(operation, a, b):

    if operation == 'omin':
        return min(a, b)

    elif operation == 'omax':
        return max(a, b)

    elif operation == 'oabs':
        return abs(a)

    elif operation == 'sha1':
        return sha1(a)

    elif operation == 'sha256':
        return sha256(a)

    elif operation == 'hash160':
        return hash160(a)

    elif operation == 'hash256':
        return hash256(a)

    return 'unknown'



from boa.interop.Neo.Runtime import GetTrigger
from boa.interop.Neo.TriggerType import Application, Verification
# from boa.builtins import concat


def Main():

    trigger = GetTrigger()

    return trigger == Application()



import pdb
import binascii
from boa.code import pyop


class appcall():

    script_hash = None
    script_args = None

    method_name = None

    def __init__(self, block):

        arguments = []

        for i, item in enumerate(block):
            if item.opcode == pyop.LOAD_CONST:
                arguments.append(item.arg)
            elif item.opcode == pyop.STORE_NAME:
                self.method_name = item.arg

        self.script_hash = arguments[0]

        self.script_args = arguments

        if type(self.script_hash) is str:
            if len(self.script_hash) != 40:
                raise Exception(
                    "Invalid script hash! length of string must be 40")
        elif type(self.script_hash) in [bytes, bytearray]:
            if len(self.script_hash) != 20:
                raise Exception(
                    "Invalid Script hash, length in bytes must be 20")
        else:
            raise Exception(
                "Invalid script hash type.  must be string, bytes, or bytearray")

    @property
    def script_hash_addr(self):
        """

        :return:
        """

        return appcall.to_script_hash_data(self.script_hash)

    @staticmethod
    def to_script_hash_data(item):
        """

        :return:
        """
        b_array = None
        if type(item) is str:
            bstring = item.encode('utf-8')
            b_array = bytearray(binascii.unhexlify(bstring))
        elif type(item) is bytearray:
            pass
        elif type(item) is bytes:
            b_array = bytearray(item)
        else:
            raise Exception("Invalid script hash")

        b_array.reverse()

        return bytes(b_array)



from boa.interop.Neo.Runtime import Notify, Serialize, Deserialize
from boa.interop.Neo.Storage import Put, Get, GetContext

ctx = GetContext()


def Main(operation):

    # create an array
    stuff = ['a', 3, ['j', 3, 5], 'jk', 'lmnopqr']

    # serialize it
    to_save = Serialize(stuff)
    Put(ctx, 'serialized', to_save)

    if operation == 1:
        return to_save

    elif operation == 2:

        to_retrieve = Get(ctx, 'serialized')
        return to_retrieve

    elif operation == 3:

        to_retrieve = Get(ctx, 'serialized')
        deserialized = Deserialize(to_retrieve)
        return deserialized

    elif operation == 4:

        to_retrieve = Get(ctx, 'serialized')
        deserialized = Deserialize(to_retrieve)
        return deserialized[2]

    return False
